#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#if defined(__APPLE__)
#include <OpenCL/OpenCL.h>
#include <sys/sysctl.h>
#elif defined(__linux)
#include <CL/cl.h>
#elif defined(_WIN32)
#include <CL/cl.h>
#endif



#define messAssert(returnStatus ) {if (!(returnStatus==CL_SUCCESS)) {fprintf(stderr, "[%s:%u] assert: failed assertion %d\n", __FILE__, __LINE__, returnStatus); abort();}}

char * load_program_source(const char *filename)
{
	struct stat statbuf;
	FILE *fh;
	char *source;

	fh = fopen(filename, "r");
  if (fh == 0) {
    fprintf(stderr, "Could not find file %s\n", filename);
		return 0;
  }

	stat(filename, &statbuf);
	source = (char *) malloc(statbuf.st_size + 1);
  memset(source, 0, statbuf.st_size + 1);
	fread(source, 1, statbuf.st_size, fh);
	source[statbuf.st_size] = '\0';

	return source;
}

void setBufferData(cl_float* hostBuffer, cl_command_queue cmd_queue, cl_mem buffer, size_t numBytes);

int main (int argc, char const *argv[])
{
  cl_program program_value;
  cl_program program_ptr;
	cl_kernel kernel_value;
  cl_kernel kernel_ptr;
	cl_command_queue cmd_queue;
	cl_context   context;
  cl_int result;
  cl_device_id device;
  cl_mem buffer;
  cl_event event;

  /* Get a platform */
  cl_platform_id platform;
  result = clGetPlatformIDs(1, &platform, NULL);
  assert( result == CL_SUCCESS );

  cl_context_properties properties[3] = {0};
  properties[0] = CL_CONTEXT_PLATFORM;
  properties[1] = (cl_context_properties)platform;


  /* Set up device and context */
  result = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
	assert(result == CL_SUCCESS);

	cl_char vendor_name[1024];
	cl_char device_name[1024];
	result = clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(vendor_name), vendor_name, NULL);
	result |= clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name), device_name, NULL);
	assert(result == CL_SUCCESS);
	printf("Setting up device %s - %s\n\n", vendor_name, device_name);

	context = clCreateContext(properties, 1, &device, NULL, NULL, &result);
	assert(result == CL_SUCCESS);

	cmd_queue = clCreateCommandQueue(context, device, 0, &result);
	assert(result == CL_SUCCESS);



	/* Load source and create kernel */
	const char * filename = "bug_kernel.cl";
	char *program_source = load_program_source(filename);
	program_value = clCreateProgramWithSource(context, 1, (const char**)&program_source, NULL, &result);
	assert(result == CL_SUCCESS);
  program_ptr= clCreateProgramWithSource(context, 1, (const char**)&program_source, NULL, &result);
  assert(result == CL_SUCCESS);

	result = clBuildProgram(program_value, 0, NULL, "-DPASS_BY_VALUE", NULL, NULL);
	if (result != CL_SUCCESS)
	{
    char buildLog[128 * 1024];
	  result = clGetProgramBuildInfo(program_value, device, CL_PROGRAM_BUILD_LOG, sizeof(buildLog), buildLog, NULL);
	  assert(result == CL_SUCCESS);
    printf("Could not build program: \n\n%s\n", buildLog);
    return -1;
	}

  result = clBuildProgram(program_ptr, 0, NULL, "-DPASS_BY_POINTER", NULL, NULL);
  if (result != CL_SUCCESS)
  {
    char buildLog[128 * 1024];
    result = clGetProgramBuildInfo(program_ptr, device, CL_PROGRAM_BUILD_LOG, sizeof(buildLog), buildLog, NULL);
    assert(result == CL_SUCCESS);
    printf("Could not build program: \n\n%s\n", buildLog);
    return -1;
  }


	kernel_value = clCreateKernel(program_value, "bug_kernel", &result);
	assert(result == CL_SUCCESS);

  kernel_ptr = clCreateKernel(program_ptr, "bug_kernel", &result);
  assert(result == CL_SUCCESS);



	/* Create buffer */
  size_t localRange[3] = {1, 1, 1};
  size_t globalRange[3] = {1, 1, 1};

  size_t numElements = 2;
  size_t numBytes = numElements * sizeof(cl_float);
  cl_float *hostBuffer = (cl_float *)malloc(numBytes);
	buffer = clCreateBuffer(context, CL_MEM_READ_WRITE, numBytes, NULL, &result);
	assert(result == CL_SUCCESS);

  


	/* Set arguments and execute pointer kernel */
  printf("Running kernel with pass-as-pointer. Works for me\n");
  
  setBufferData(hostBuffer, cmd_queue, buffer, numBytes);
  result  = clSetKernelArg(kernel_ptr,  0, sizeof(cl_mem), &buffer);
  messAssert(result);
  
  result = clEnqueueNDRangeKernel(cmd_queue, kernel_ptr, 3, NULL, globalRange, localRange, 0, NULL, &event);
  messAssert(result);
  result = clFinish(cmd_queue);
  messAssert(result);



  

	result = clEnqueueReadBuffer(cmd_queue, buffer, CL_TRUE, 0, numBytes, hostBuffer, 0, NULL, NULL);
	assert(result == CL_SUCCESS);

  printf("Expecting a list of 2 floats, both being 99.0:\n");
  printf(" [%d]: %f\n", 0, hostBuffer[0]);
  printf(" [%d]: %f\n", 1, hostBuffer[1]);



	/* Set arguments and execute value kernel */
  printf("Running kernel with pass-as-value. Crashes for me\n");

  setBufferData(hostBuffer, cmd_queue, buffer, numBytes);
  result  = clSetKernelArg(kernel_value,  0, sizeof(cl_mem), &buffer);
  messAssert(result);

  
  result = clEnqueueNDRangeKernel(cmd_queue, kernel_value, 3, NULL, globalRange, localRange, 0, NULL, &event);
  messAssert(result);
  result = clFinish(cmd_queue);
  messAssert(result);

  result = clEnqueueReadBuffer(cmd_queue, buffer, CL_TRUE, 0, numBytes, hostBuffer, 0, NULL, NULL);
  assert(result == CL_SUCCESS);

  printf("Expecting a list of 2 floats, both being 99.0:\n");
  printf(" [%d]: %f\n", 0, hostBuffer[0]);
  printf(" [%d]: %f\n", 1, hostBuffer[1]);


	clReleaseMemObject(buffer);
	clReleaseCommandQueue(cmd_queue);
	clReleaseContext(context);

  return 0;
}



void setBufferData(cl_float* hostBuffer, cl_command_queue cmd_queue, cl_mem buffer, size_t numBytes) {
  hostBuffer[0] = 0.0f;
  hostBuffer[1] = 99.0f;

  cl_int result = clEnqueueWriteBuffer(cmd_queue, buffer, CL_TRUE, 0, numBytes, hostBuffer, 0, NULL, NULL);
  messAssert(result);
}